import { createElement, useEffect, useState } from 'react';
import { isObservable } from 'rxjs';
import { distinctUntilChanged, takeUntil } from 'rxjs/operators';
import { $ } from './fragment';
import { useDestroyObservable, useEmptyObject } from './shared';
export function createElement$(element) {
    return function element$(props) {
        // state for renderable props
        var _a = useState(Object.create(null)), stateProps = _a[0], setStateProps = _a[1];
        // placeholder for Observable.error case -- will use it to rethrow error
        var _b = useState(null), error = _b[0], setError = _b[1];
        var destroy$ = useDestroyObservable();
        // store prev props to compare
        var _prevProps = useEmptyObject();
        // keep subscriptions to unsubscribe on dynamic prop update
        var _subs = useEmptyObject();
        useEffect(function () {
            var isDirty = false;
            // check for obsolete props
            var delProps = Object.create(null);
            Object.keys(_prevProps).forEach(function (key) {
                if (key in props) {
                    return;
                }
                isDirty = true;
                delete _prevProps[key];
                // if previous property was Observable
                // kill subscription
                cleanSubscription(_subs, key);
                // remove from static props
                delProps[key] = void 0;
            });
            // update/track new props
            var nextProps = Object.create(null);
            var streamKeys = [];
            Object.keys(props).forEach(function (key) {
                // children are covered via <$>
                if (key == 'children') {
                    return;
                }
                var value = props[key];
                var prevValue = _prevProps[key];
                var equal = Object.is(value, prevValue);
                if (equal) {
                    return;
                }
                isDirty = true;
                _prevProps[key] = value;
                // if property changes and previous was Observable
                // we need to kill subscription
                cleanSubscription(_subs, key);
                // observable input params are auto observed
                if (isObservable(value)) {
                    nextProps[key] = void 0;
                    streamKeys.push(key);
                }
                // all static props are directly updated
                else {
                    nextProps[key] = value;
                }
            });
            // remove obsolete props
            // & update static props
            if (isDirty) {
                setStateProps(function (p) { return Object.assign({}, p, delProps, nextProps); });
            }
            // subscribe to new streams
            streamKeys.forEach(function (key) {
                _subs[key] = props[key]
                    .pipe(distinctUntilChanged(), takeUntil(destroy$))
                    .subscribe({
                    // on value updates -- update props
                    next: function (data) {
                        setStateProps(function (p) {
                            var _a;
                            return Object.assign({}, p, (_a = {}, _a[key] = data, _a));
                        });
                    },
                    // on error -- rethrow error
                    error: function (error) {
                        setError({ error: error });
                    }
                    // on complete we just keep using accumulated value
                });
            });
        }, [props]);
        // if error -- throw
        if (error) {
            throw error.error;
        }
        return createElement(element, stateProps, 
        // if children exist
        // they might be observable
        // so we pass em to <$> fragment
        // NOTE: children might not exist for elements like <input />
        props.children
            ? createElement($, null, props.children)
            : null);
    };
}
// helpers
function cleanSubscription(store, key) {
    if (store[key]) {
        store[key].unsubscribe();
        delete store[key];
    }
}
